from PyQt5.QtWidgets import QWidget, QFrame
from PyQt5.Qt import Qt
from .TextUpdate import TextUpdateBase
class BooleanLED(QFrame, TextUpdateBase):
	def setColor(self, color):
		p = self.palette()
		p.setColor(self.backgroundRole(), color)
		self.setPalette(p)
	def setVal(self, val):
		if val:
			self.setColor(self.onColor)
		else:
			self.setColor(self.offColor)
	def __init__(self, parent=None):
		super().__init__(parent)
		p = self.palette()
		self.offColor = Qt.black
		self.onColor = Qt.green
		self.setVal(False)
		self.setAutoFillBackground(True)
		#self.setStyleSheet("border: 1px solid black;")
	def default_code(self):
		return """
		from PyQt5.Qt import Qt
		from bsstudio.functions import widgetValue
		self.offColor = Qt.black
		self.onColor = Qt.green
		ui = self.ui
		if self.source != "":
			b = widgetValue(eval(self.source))
			self.setVal(b)
		"""[1:]



from . TextUpdate import TextUpdateBase
from PyQt5.QtWidgets import QProgressBar
from PyQt5.QtCore import Q_ENUMS
from .REButton import makeProperty
import math
from enum import Enum

class ScaleType:
	Linear = 1
	Logarithmic = 2

class BarUpdate(TextUpdateBase, QProgressBar, ScaleType):
	Q_ENUMS(ScaleType)
	ScaleType=ScaleType
	def __init__(self, parent):
		QProgressBar.__init__(self, parent)
		TextUpdateBase.__init__(self, parent)
		self._scientificNotation = False
		self._scale = ScaleType.Linear
		self.N = 100
		self.setMinimum(0)
		self.setMaximum(self.N)
		self._minimumValue = "0"
		self._maximumValue = "100"
	def setBarValue(self, val):
		low = eval(self._minimumValue)
		high = eval(self._maximumValue)
		barPercent = (val - low)/(high - low)
		self.setValue(barPercent*self.N)
	def setBarValueLog(self, val):
		low = math.log10(eval(self._minimumValue))
		high = math.log10(eval(self._maximumValue))
		val = math.log10(val)
		barPercent = (val - low)/(high - low)
		self.setValue(barPercent*self.N)
	def default_code(self):
		return """
			ui = self.ui
			from bsstudio.functions import widgetValue
			value = widgetValue(eval(self.source))
			if self.scale == self.ScaleType.Linear:
				self.setBarValue(value)
			elif self.scale == self.ScaleType.Logarithmic:
				self.setBarValueLog(value)
			if self.scientificNotation:
				self.setFormat("%.02E" % value) 
			else:
				self.setFormat("%.02f" % value) 
			"""[:-1]

	minimumValue = makeProperty("minimumValue")
	maximumValue = makeProperty("maximumValue")
	scientificNotation = makeProperty("scientificNotation", bool)
	scale = makeProperty("scale", ScaleType)
	



from PyQt5 import QtDesigner, QtGui, QtWidgets, QtCore
#from qtpy.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton, QPlainTextEdit
#from qtpy.QtDesigner import QExtensionFactory
from PyQt5.QtDesigner import QExtensionFactory
from PyQt5.QtCore import pyqtProperty as Property
from ophyd.sim import det
import inspect
from itertools import dropwhile
import textwrap
from .Base import BaseWidget
import sys
from IPython import get_ipython
from PyQt5 import QtCore
from ..worker import Worker, WorkerSignals
from functools import partial
import logging
import time

logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.WARN)


class CodeObject(BaseWidget):

	def __init__(self, parent=None, copyNameSpace=True):
		self.parent = parent
		super().__init__(parent)
		code = textwrap.dedent(self.default_code())
		self._code = bytes(code, "utf-8")
		self._paused = True
		self.threadpool = QtCore.QThreadPool(self)
		self._useThreading = False
		self._copyNameSpace = copyNameSpace
		self.ns_extras = {}
		#self.worker = None

	def addToNameSpace(self, key, val):
		self.ns_extras[key] = val

	@Property("QByteArray", designable=True)
	def code(self):
		return self._code

	@code.setter
	def code(self, val):
		self._code = val
	
	def default_code(self):
		return ""

	def pause_widget(self):
		logger.info("pausing widget")
		self._paused = True

	def resume_widget(self):
		logger.info("unpausing widget")
		self._paused = False
		#self.setup_namespace()

		
	def setup_namespace(self):
		t0 = time.time()
		ip = get_ipython()
		if self._copyNameSpace:
			ns = ip.user_ns.copy()
		else:
			ns = ip.user_ns
			
		ns['self'] = self
		ns.update(self.ns_extras)
		self.ns = ns
		logger.info("setup_namespace duration for "+self.objectName()+": "+str(time.time()-t0))

	def runInNameSpace(self, codeString):
		if self._paused:
			logger.info("widget paused")
			return
		#ns = vars(sys.modules[self.__class__.__module__])
		self.setup_namespace()
		logger.info("runInNameSpace for "+self.objectName())
		
		try:
			#exec(self._code, ns)
			t0 = time.time()
			exec(codeString, self.ns)
			logger.info("exec duration for "+self.objectName()+": "+str(time.time()-t0))
		except BaseException as e:
			additional_info = " Check code in "+self.objectName()+" widget"
			raise type(e)(str(e) + additional_info).with_traceback(sys.exc_info()[2])


	def runPaused(self):
		pass

	def runCode(self):
		if self._paused:
			self.runPaused()
			return
		if not self._useThreading:
			self.runInNameSpace(self._code)
		else:
			logger.info("Active thread count: "+str(self.threadpool.activeThreadCount()))
			logger.info("Max thread count: "+str(self.threadpool.maxThreadCount()))
			logger.info("No other threads running: "+str(self.threadpool.waitForDone(0)))
			#self.worker = Worker(partial(self.runInNameSpace, self._code))
			#self.threadpool.start(self.worker)
			worker = Worker(partial(self.runInNameSpace, self._code))
			self.threadpool.start(worker)



import re

def find_right_paren(s, start=0):
	m = re.search("(^|[^\\\])[)]",s[start:])
	if m is None:
		return None
	return start + m.end() - 1

def get_first_outer_macro(s):
	left = s.find("$(")
	left2 = s.find("$(", left+1)
	#right = find_right_paren(s, 0)
	right = find_right_paren(s, left)
	while right is not None and left2<right and left2!=-1:
		gfom = get_first_outer_macro(s[left2:])
		n = left2+len(gfom)
		right = find_right_paren(s, n)
		left2 = s.find("$(", left2+len(gfom))
	if right is None:
		return None
	return s[left:right+1]

def call_until_invariant(f, args):
	x = args[0]
	if f(*args)==x:
		return x
	args[0] = f(*args)
	return call_until_invariant(f,args)

def macro_name(s):
	gfom = get_first_outer_macro(s)
	if gfom is None:
		return None
	return gfom[2:-1].split("=")[0]

def get_macro_by_name(s, name):
	i = s.find("$(")
	while(i!=-1):
		if macro_name(s[i:]) == name:
			return get_first_outer_macro(s[i:])
		i = s.find("$(",i+1)
	return None

def subst_str_once(s, macros):
	for m_k in macros.keys():
		if get_macro_by_name(s, m_k) is not None:
			s = s.replace(get_macro_by_name(s, m_k), macros[m_k])
	return s.replace("\)",")")


def subst_str(s, macros):
	return call_until_invariant(subst_str_once, [s, macros])

def get_first_outer_macro_default(s):
	m = get_first_outer_macro(s)
	if m is None:
		return None
	ms = m.split("=")
	if len(ms)<2:
		return get_first_outer_macro_default(s[s.find(m)+1:])
	return m


def subst_str_defaults_once(s):
	m = get_first_outer_macro_default(s)
	if m is None:
		return s
	ms = m.split("=")
	v = ms[1][:-1]
	return s.replace(m, v)

def subst_str_defaults(s):
	return call_until_invariant(subst_str_defaults_once, [s])

def subst_str_all(s, macros):
	s = subst_str(s, macros)
	return subst_str_defaults(s)




from .CodeObject import CodeObject
from .TextUpdate import TextUpdateBase
from .REButton import makeProperty
from PyQt5.QtWidgets import QLineEdit
import textwrap
import logging

logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.WARN)

class LineInput(QLineEdit, TextUpdateBase):
	def __init__(self, parent=None,*, sig=""):
		#super().__init__(parent)
		QLineEdit.__init__(self,parent)
		TextUpdateBase.__init__(self,parent)
		self._copyNameSpace = False
		#CodeObject.__init__(self,parent,copyNameSpace=False)
		self._destination = ""
		self.destination = sig
		textUpdateCode = textwrap.dedent(TextUpdateBase.default_code(self))
		self._textUpdateCode = bytes(textUpdateCode, "utf-8")
		self.returnPressed.connect(self.runCode)

	def timeout(self):
		if not self.hasFocus() and self.source!="":
			#try:
			#	self.runInNameSpace(self.textUpdateCode)
			#except:
			#	self.setText("unknown")
			logger.info(self.objectName()+":timeout")
			logger.info(self.objectName()+":code:"+str(self.textUpdateCode))
			self.runInNameSpace(self.textUpdateCode)

	def default_code(self):
		return """
			from bsstudio.functions import widgetValue
			ui = self.ui
			#destination = widgetValue(eval(self.destination))
			#val = widgetValue(eval(self.text()))
			if self.hasFocus() and self.destination!="":
				exec(self.destination+" = "+self.text())
			"""[1:]
		
	textUpdateCode = makeProperty("textUpdateCode", "QByteArray")
	destination = makeProperty("destination")



# Imports
from PyQt5 import QtWidgets, QtCore
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as Canvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.backends.backend_qt5 import StatusbarQt
from matplotlib.backend_bases import StatusbarBase
import matplotlib
from .Base import BaseWidget

# Ensure using PyQt5 backend
matplotlib.use('QT5Agg')

#def on_mouse_move(event):
#	print('Event received:',event.x,event.y)
# Matplotlib canvas class to create figure
class MplCanvas(Canvas):
	def __init__(self):
		self.fig = Figure()
		self.ax = self.fig.add_subplot(111)

		Canvas.__init__(self, self.fig)
		Canvas.setSizePolicy(self, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
		Canvas.updateGeometry(self)
		#self.mpl_connect('motion_notify_event',on_mouse_move)
		#self.ax.plot([0, 1], [1, 1])

	def wipe(self):
		self.fig.clf()
		self.ax = self.fig.add_subplot(111)
		

# Matplotlib widget
class MplWidget(QtWidgets.QWidget, BaseWidget):
	def __init__(self, parent=None):
		QtWidgets.QWidget.__init__(self, parent)   # Inherit from QWidget
		self.canvas = MplCanvas()				   # Create canvas object

		self.toolbar = NavigationToolbar(self.canvas, self)
		#self.statusbar = StatusbarBase(self.canvas.manager.toolmanager)
		#self.statusbar = StatusbarQt(self, None)

		self.vbl = QtWidgets.QVBoxLayout()		   # Set box for plotting
		self.vbl.addWidget(self.toolbar)
		self.vbl.addWidget(self.canvas)
		#self.vbl.addWidget(self.statusbar)
		#self.vbl.addWidget(QtCore.Qt.BottomToolBarArea, NavigationToolbar(self.canvas, self))
		self.setLayout(self.vbl)
		def format_coord(x, y):
			return "x={:0.2f},y={:0.2f}".format(x,y)
		self.canvas.ax.format_coord = format_coord
		#self.canvas.ax.figure.tight_layout()
		#self.canvas.window().statusbar().setVisible(True)

	def sizeHint(self):
		return QtCore.QSize(400, 300) 



from .TextUpdate import TextUpdateBase
from .mplwidget import MplWidget
from .REButton import makeProperty
from PyQt5 import QtCore
from PyQt5.QtWidgets import QVBoxLayout
import pyqtgraph as pg
import matplotlib.cm
import time
from functools import partial

import logging

logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.WARN)
pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')

from ..worker import Worker, WorkerSignals
class ArrayImage(TextUpdateBase, pg.GraphicsLayoutWidget):
	def __init__(self, parent):
		#super().__init__(parent)
		pg.GraphicsLayoutWidget.__init__(self, parent)
		TextUpdateBase.__init__(self, parent)
		self._updatePeriod = "10000"
		self._enableHistogram = False
		self.updatePeriod_ = eval(self._updatePeriod)
		#self.threadpool.setMaxThreadCount(1)
		self.threadType = "qthread"
		#self.ui.histogram.hide()
		#self.ui.roiBtn.hide()
		#self.ui.menuBtn.hide()
		self.imv = pg.ImageItem()
		#self.plot = pg.PlotItem()
		#self.view = self.addViewBox()
		self.view = self.addPlot()
		#self.view.invertY(True)
		#self.view.addItem(self.plot)
		#self.view.setBorder(None)
		#self.centralWidget.setBorder((0,0,0))
		self.centralWidget.layout.setSpacing(0)
		self.centralWidget.setContentsMargins(0,0,0,0)
		self.centralWidget.layout.setContentsMargins(0,0,0,0)
		self.view.setContentsMargins(0,0,0,0)
		print(self.centralWidget)
		self.view.addItem(self.imv)
		self.lineh = self.view.addLine(y=.5)
		self.lineh.setMovable(True)
		self.linev = self.view.addLine(x=.5)
		self.linev.setMovable(True)
		self.linesToggle()
		#self.plot.addItem(self.imv)
		self.hist = pg.HistogramLUTItem(image=self.imv,fillHistogram=True)
		#self.hist.setLevels(0,256)
		self.setLayout(QVBoxLayout())
		self.addItem(self.hist)
		#pg.GraphicsView.setCentralItem(self, self.imv)
		histogramAction = self.view.getViewBox().menu.addAction("Histogram")
		linesAction = self.view.getViewBox().menu.addAction("Lines")
		histogramAction.triggered.connect(self.histogramToggle)
		linesAction.triggered.connect(self.linesToggle)
		self.hist.hide()

		vb = self.view.getViewBox()

		submenu_cmaps = vb.menu.addMenu('cmaps');
		col_inferno = submenu_cmaps.addAction('inferno')
		col_viridis = submenu_cmaps.addAction('viridis')
		col_cividis = submenu_cmaps.addAction('cividis')
		col_magma = submenu_cmaps.addAction('magma')
		col_gray = submenu_cmaps.addAction('gray')

		def set_cmap(c):
			import bsstudio.widgets.ut as ut
			pos, rgba_colors = zip(*ut.cmapToColormap(c))
			# Set the colormap
			pgColormap = pg.ColorMap(pos, rgba_colors)
			self.imv.setLookupTable(pgColormap.getLookupTable())
			self.hist.gradient.setColorMap(pgColormap)
		col_gray.triggered.connect(partial(set_cmap, matplotlib.cm.gray))
		col_viridis.triggered.connect(partial(set_cmap, matplotlib.cm.viridis))
		col_magma.triggered.connect(partial(set_cmap, matplotlib.cm.magma))
		col_cividis.triggered.connect(partial(set_cmap, matplotlib.cm.cividis))
		col_inferno.triggered.connect(partial(set_cmap, matplotlib.cm.inferno))

		vb.setMouseMode(vb.RectMode)
		#self.view.getViewBox().setBackgroundColor("w")
		#self.setBackground("w")


		
	def toggleWidgetVisibility(self, w):
		if w.isVisible():
			w.hide()
		else:
			w.show()

	def histogramToggle(self):
		"""
		if self.hist.isVisible():
			self.hist.hide()
		else:
			self.hist.show()
		"""
		self.toggleWidgetVisibility(self.hist)

	def linesToggle(self):
		self.toggleWidgetVisibility(self.lineh)
		self.toggleWidgetVisibility(self.linev)
	
	def setUpdatePeriod(self, p):
		self.updatePeriod_ = p

	def default_code(self):
		return """
		import logging
		logger = logging.getLogger(__name__)
		import time
		t0 = time.time()
		from PyQt5 import QtCore
		from bsstudio.functions import widgetValue
		from bsstudio.worker import Worker, WorkerSignals
		import numpy as np
		import pyqtgraph as pg
		pg.setConfigOption('background', 'w')
		pg.setConfigOption('foreground', 'k')
		#if not self.enableHistogram:
		#	self.hist.hide()
		#self.canvas.ax.clear()
		array = None
		logger.info("time before eval source: "+str(time.time()-t0))
		if self.source != "":
			array = eval(self.source)
		logger.info("time before widgetValue: "+str(time.time()-t0))
		array = widgetValue(array)
		t2 = time.time()
		logger.info("time before imshow: "+str(t2-t0))
		#self.canvas.ax.imshow(array)
		if not hasattr(self,"ran_once"):
			self.imv.setImage(array,autoRange=True, autoLevels=True)
			self.hist.setHistogramRange(self.imv.levels[0],self.imv.levels[1],padding=0)
			#self.view.autoRange(padding=0)
		else:
			self.imv.setImage(array,autoRange=False, autoLevels=False)
		t3 = time.time()
		logger.info("time after imshow: "+str(t3-t0))
		#self.canvas.draw()
		logger.info("time after draw: "+str(time.time()-t0))
		self.setUpdatePeriod(eval(self.updatePeriod))
		t1 = time.time()
		logger.info("time at end of runCode: "+str(t1-t0))
		self.ran_once = True
		"""[1:]

	def closeEvent(self, evt):
		self.pause_widget()
		self.worker.cancel()
		pg.GraphicsLayoutWidget.closeEvent(self,evt)
		TextUpdateBase.closeEvent(self,evt)

	def resume_widget(self):
		TextUpdateBase.resume_widget(self)

	#enableHistogram = makeProperty("enableHistogram", bool)




from .CodeObject import CodeObject
from .REButton import makeProperty
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QSizePolicy
from PyQt5 import QtCore
from PyQt5.Qt import Qt
from .embedframe import CodeContainer
from .TextUpdate import TextUpdate
from .lineinput import LineInput
from ..functions import widgetValueString
import logging
import time
		
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class OphydProperties(CodeContainer):
	ophydObject = makeProperty("ophydObject")
	
	def __init__(self, parent):
		super().__init__(parent)
		self._ophydObject = ""
		#self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

	def makeRow(self, obj, name, obj_name):
		logger.debug("here1")
		sig = getattr(obj,name)
		hlayout = QHBoxLayout()
		hlayout.setAlignment(Qt.AlignLeft)
		w = QLabel(name)
		logger.debug("here2")
		#w.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
		w.setFixedHeight(25)
		w.setFixedWidth(200)
		hlayout.addWidget(w)
		#w = TextUpdate(self, sig=obj_name+"."+name+".value")
		if hasattr(sig, "value"):
			logger.info("hasattr value:"+obj.name)
			#w = TextUpdate(self, sig=obj_name+"."+name+".get(timeout=.5)")
			w = TextUpdate(self, sig=obj_name+"."+name+".get()")
		else:
			logger.info("does not hasattr value:"+obj.name)
			w = TextUpdate(self, sig="str('unknown')")
		#w.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
		logger.debug("here3")
		w.setFixedHeight(25)
		w.setFixedWidth(200)
		hlayout.addWidget(w)
		logger.debug("here")
		if not hasattr(sig, "write_access") or sig.write_access:
			#w = LineInput(self, sig=obj_name+"."+name+".value")
			w = LineInput(self, sig=obj_name+"."+name+".put(eval(self.text())) #")
			w.setFixedWidth(100)
			hlayout.addWidget(w)
		return hlayout

		
	def createFields(self, obj, obj_name):
		print("create fields")
		layout = QVBoxLayout()
		hlayout = QHBoxLayout()
		hlayout.setAlignment(Qt.AlignLeft)
		title = QLabel("<b>"+obj.name+"</b>")
		title.setFixedHeight(25)
		title.setFixedWidth(100)
	
		hlayout.addWidget(title)
		layout.addLayout(hlayout)
		for name in obj.component_names:
			try:
				hlayout = self.makeRow(obj, name, obj_name)
				layout.addLayout(hlayout)
			except:
				logger.warn("error making row for " + obj_name + " and " + name)
		self.setLayout(layout)
			
	def resume_widget(self):
		self._paused = False
		self.runCode()
		self.resume_children()

	def default_code(self):
		return """
			from bsstudio.functions import widgetValueString
			ui = self.ui
			obj_name = widgetValueString(self, self.ophydObject) 
			ophydObject = eval(obj_name)
			self.createFields(ophydObject, obj_name)
			"""[1:]



import matplotlib.cm
import numpy as np

#######################################################################################
################## COLOR MAPS ########################################################
#######################################################################################
"""
This is an example how to use an ImageView with Matplotlib Colormaps (cmap).
The function 'cmapToColormap' converts the Matplotlib format to the internal
format of PyQtGraph that is used in the GradientEditorItem. The function
itself has no dependencies on Matplotlib! Hence the weird if clauses with
'hasattr' instead of 'isinstance'.
"""


import collections


def cmapToColormap(cmap, nTicks=16):
    """
    Converts a Matplotlib cmap to pyqtgraphs colormaps. No dependency on matplotlib.
    Parameters:
    *cmap*: Cmap object. Imported from matplotlib.cm.*
    *nTicks*: Number of ticks to create when dict of functions is used. Otherwise unused.
    """

    # Case #1: a dictionary with 'red'/'green'/'blue' values as list of ranges (e.g. 'jet')
    # The parameter 'cmap' is a 'matplotlib.colors.LinearSegmentedColormap' instance ...
    if hasattr(cmap, '_segmentdata'):
        colordata = getattr(cmap, '_segmentdata')
        if ('red' in colordata) and isinstance(colordata['red'], collections.Sequence):

            # collect the color ranges from all channels into one dict to get unique indices
            posDict = {}
            for idx, channel in enumerate(('red', 'green', 'blue')):
                for colorRange in colordata[channel]:
                    posDict.setdefault(colorRange[0], [-1, -1, -1])[idx] = colorRange[2]

            indexList = list(posDict.keys())
            indexList.sort()
            # interpolate missing values (== -1)
            for channel in range(3):  # R,G,B
                startIdx = indexList[0]
                emptyIdx = []
                for curIdx in indexList:
                    if posDict[curIdx][channel] == -1:
                        emptyIdx.append(curIdx)
                    elif curIdx != indexList[0]:
                        for eIdx in emptyIdx:
                            rPos = (eIdx - startIdx) / (curIdx - startIdx)
                            vStart = posDict[startIdx][channel]
                            vRange = (posDict[curIdx][channel] - posDict[startIdx][channel])
                            posDict[eIdx][channel] = rPos * vRange + vStart
                        startIdx = curIdx
                        del emptyIdx[:]
            for channel in range(3):  # R,G,B
                for curIdx in indexList:
                    posDict[curIdx][channel] *= 255

            rgb_list = [[i, posDict[i]] for i in indexList]

        # Case #2: a dictionary with 'red'/'green'/'blue' values as functions (e.g. 'gnuplot')
        elif ('red' in colordata) and isinstance(colordata['red'], collections.Callable):
            indices = np.linspace(0., 1., nTicks)
            luts = [np.clip(np.array(colordata[rgb](indices), dtype=np.float), 0, 1) * 255 \
                    for rgb in ('red', 'green', 'blue')]
            rgb_list = zip(indices, list(zip(*luts)))

    # If the parameter 'cmap' is a 'matplotlib.colors.ListedColormap' instance, with the attributes 'colors' and 'N'
    elif hasattr(cmap, 'colors') and hasattr(cmap, 'N'):
        colordata = getattr(cmap, 'colors')
        # Case #3: a list with RGB values (e.g. 'seismic')
        if len(colordata[0]) == 3:
            indices = np.linspace(0., 1., len(colordata))
            scaledRgbTuples = [(rgbTuple[0] * 255, rgbTuple[1] * 255, rgbTuple[2] * 255) for rgbTuple in colordata]
            rgb_list = zip(indices, scaledRgbTuples)

        # Case #4: a list of tuples with positions and RGB-values (e.g. 'terrain')
        # -> this section is probably not needed anymore!?
        elif len(colordata[0]) == 2:
            rgb_list = [(idx, (vals[0] * 255, vals[1] * 255, vals[2] * 255)) for idx, vals in colordata]

    # Case #X: unknown format or datatype was the wrong object type
    else:
        raise ValueError("[cmapToColormap] Unknown cmap format or not a cmap!")

    # Convert the RGB float values to RGBA integer values
    return list([(pos, (int(r), int(g), int(b), 255)) for pos, (r, g, b) in rgb_list])
#############################





from .window import getMainWindow, getWidgetById, isMainWindow
from .window import load
#from .window import ui
from .test import abc
import logging
#logging.basicConfig(level=logging.WARN, format='%(message)s')
#logging.basicConfig(filename="log", filemode='a', level=logging.WARN, format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", datefmt='%Y-%m-%d %H:%M:%S')
#logging.getLogger().addHandler(logging.StreamHandler())
#logging.getLogger().addHandler(logging.StreamHandler())



from PyQt5 import QtDesigner, QtGui, QtWidgets, QtCore
#from qtpy.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton, QPlainTextEdit
#from qtpy.QtDesigner import QExtensionFactory
from PyQt5.QtDesigner import QExtensionFactory
from PyQt5.QtCore import pyqtProperty as Property
from PyQt5.QtCore import pyqtSignal
from ophyd.sim import det
from ..functions import getTopObject
import inspect
from itertools import dropwhile
import textwrap
import time
import logging

logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.WARN)

all_bss_widgets = []

class BaseWidget:
	signal = pyqtSignal()

	def __init__(self, parent=None):
		self.parent = parent
		global all_bss_widgets
		all_bss_widgets.append(self)
		self.isTopLevel = False

	def initialize(self):
		pass

	def pause_widget(self):
		pass

	def resume_widget(self):
		pass


	def windowFileName(self):
		fileName = None
		if not self._paused:
			fileName = getTopObject(self).uiFilePath
		else:
			try:
				fileName = self.core.formWindowManager().activeFormWindow().fileName()
			except:
				logger.info("Error getting window filename in designer")
		return fileName


	def ui(self):
		from ..functions import makeUiFunction
		_ui = makeUiFunction(self)
		return _ui()

	def resume_children(self):
		children = self.findChildren(BaseWidget)
		for c in children:
			c.resume_widget()

	def closeEvent(self, evt):
		logger.info("close Event")
		self.hide()
		#self.worker.cancel()
		for child in self.findChildren(QWidget):
			try:
				child.close()
			except:
				None
		#super().closeEvent(self)





from PyQt5 import QtDesigner, QtGui, QtWidgets, QtCore
#from qtpy.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton, QPlainTextEdit, QWidget, QAction
#from qtpy.QtDesigner import QExtensionFactory
from PyQt5.QtDesigner import QExtensionFactory, QDesignerPropertyEditorInterface 
from PyQt5.QtCore import pyqtProperty as Property
from PyQt5.QtCore import QVariant, QCoreApplication
from PyQt5.QtCore import pyqtSlot, QObject
from ophyd.sim import det
import inspect
from itertools import dropwhile
import textwrap
from .widgets import REButton, RECustomPlan, CodeButton, TextUpdate, MplWidget, Scan1DButton, EmbedFrame, LineInput
from .widgets import Base
from .widgets import CodeContainer
from .widgets import DataBrowser
from .widgets import OphydProperties
from .widgets import OpenWindowButton
from .widgets import ArrayImage
from .widgets import BooleanLED
from .widgets import BarUpdate
from PyQt5.QtDesigner import QPyDesignerTaskMenuExtension 
from PyQt5.Qt import Qt
from PyQt5.QtWidgets import QWidget
from PyQt5.QtCore import QObject
from PyQt5.QtWidgets import QTreeView, QFileSystemModel
import inspect
import bsstudio
from PyQt5.QtCore import QFile
from PyQt5.Qt import Qt
from PyQt5.QtDesigner import QDesignerFormWindowInterface
import time
from . import git


#splash = QLabel("test")
#splash.setWindowFlags(Qt.SplashScreen)
#splash.show()
#splash.update()
#splash.repaint()
#time.sleep(60)

def clearLayout(layout):
	while layout.count():
		child = layout.takeAt(0)
		try:
			child.widget().hide()
			#if child.widget():
			#	child.widget().hide()
			#child.hide()
		except:
			None
		#if child.widget():
		#	child.widget().deleteLater()

global core

class EditTemplateMenuEntry(QPyDesignerTaskMenuExtension):

	def action(self):
		from PyQt5.QtWidgets import QMessageBox
		messageBox = QMessageBox(None)
		messageBox.setText("Coming soon...")
		self.messageBox = messageBox
		#self.messageBox.show()
		self.open_template()

	def open_template(self):
		#filename = "/home/bsobhani/bsw/bss_test83.ui"
		filename = self.widget.getAbsPath()
		#print(filename)
		file = QFile(filename)
		file.open(QFile.ReadWrite)
		#core.formWindowManager().activeFormWindow().setContents(file)
		#core.formWindowManager().addFormWindow(core.formWindowManager().createFormWindow())
		orig = core.formWindowManager().activeFormWindow()
		self.p = core.formWindowManager().activeFormWindow().parent()
		#self.p = QWidget()
		self.win = core.formWindowManager().createFormWindow(self.p,Qt.Widget)
		#print("win width", self.win.width())
		#print("win height", self.win.height())
		self.win.setContents(file)
		#print(self.win.contents())
		#print("win width", self.win.geometry().width())
		#print("win height", self.win.height())
		self.win.setHidden(False)
		#self.win.setVisible(True)
		self.win.setFileName(filename)
		self.win.show()
		
		core.formWindowManager().addFormWindow(self.win)
		core.formWindowManager().setActiveFormWindow(self.win)
		main = core.actionEditor().parent().parent().parent()
		mdiArea = main.children()[5]
		from PyQt5.QtWidgets import QMdiSubWindow
		self.subwindow = QMdiSubWindow(mdiArea)
		self.subwindow.setWidget(self.win)
		self.subwindow.show()

	def __init__(self, widget, parent):

		QPyDesignerTaskMenuExtension.__init__(self, parent)

		self.widget = widget
		self.editStateAction = QAction(self.tr("Edit template..."), self)
		self.editStateAction.triggered.connect(self.action)

	def preferredEditAction(self):
		return self.editStateAction

	def taskActions(self):
		return [self.editStateAction]

class EditCodeMenuEntry(QPyDesignerTaskMenuExtension):

  def __init__(self, widget, parent):

      QPyDesignerTaskMenuExtension.__init__(self, parent)

      self.widget = widget
      self.editStateAction = QAction(
          self.tr("Edit code..."), self)
      #self.connect(self.editStateAction,
      #    SIGNAL("triggered()"), self.updateLocation)

  def preferredEditAction(self):
      return self.editStateAction

  def taskActions(self):
      return [self.editStateAction]


class EditCodeTaskMenuFactory(QExtensionFactory):

  def __init__(self, parent = None):

      QExtensionFactory.__init__(self, parent)
      #print("factory...", self.createExtension)

  def createExtension(self, obj, iid, parent):
      #print("create extensions...")

      if iid != "org.qt-project.Qt.Designer.TaskMenu":
          return None

      if isinstance(obj, CodeButton):
          #print("wrhwrt")
          return EditCodeMenuEntry(obj, parent)

      if isinstance(obj, EmbedFrame):
          pass
          #return EditTemplateMenuEntry(obj, parent)
			

      return None

def debug_prompt(vars):
	while(True):
		cmd = input()
		if cmd == "q":
			break
		try:
			exec(cmd,vars)
		except:
			print("unable to run command")


core_initialized = False
def plugin_factory(cls, is_container=False):
	class PyBSPlugin(QtDesigner.QPyDesignerCustomWidgetPlugin):
		def __init__(self, cls):
			QtDesigner.QPyDesignerCustomWidgetPlugin.__init__(self)
			#print("init")
			self.cls = cls
			self.initialized = False

		def name(self):
			return self.cls.__name__

		def group(self):
			return "BSStudio Widgets"

		def isContainer(self):
			return is_container

		def icon(self):
			return QtGui.QIcon()

		def toolTip(self):
			return "test"

		def includeFile(self):
			return self.cls.__module__

		def whatsThis(self):
			return "this"

		def createWidget(self, parent):
			w = self.cls(parent)
			w.core = self.core
			w.pause_widget()
			return w

		
		def init_core(self):
			global core_initialized
			#print("core initialized", core_initialized)
			if core_initialized:
				return
			global core
			core = self.core
			children = core.formWindowManager().children()
			a = QAction("zzz",core.formWindowManager())

			#print("lllllllllgasdf")
			#print(core)
			#print(dir(core.formWindowManager()))
			app = QApplication.instance()
			app.setApplicationDisplayName("BS Studio")
			from PyQt5.QtWidgets import QLabel
			#self.asdf = QDockWidget()
			#self.asdf.show()
			#debug_prompt(locals())
			
			def try_until_successful(func):
				import threading
				thread = threading.Thread()
				def run_func():
					successful = False
					while not successful:
						try:
							func()
							successful = True
						except:
							None
				thread.run = run_func
				thread.start()
				#run_func()

			def make_git_context_menu():
				self.main = core.actionEditor().parent().parent().parent()
				self.menu = self.main.menuWidget().addMenu("aaaaaaaaa")
				git.make_menu(self.menu)
				#mainThread = QCoreApplication.instance().thread()
				#self.menu.moveToThread(mainThread)
				return self.menu

			def make_git():
				#self.asdf = QLabel("asdf")
				self.asdf = QTreeView()
				self.asdf.model = QFileSystemModel()
				self.asdf.model.setRootPath('')
				self.asdf.setModel(self.asdf.model)
				from PyQt5.QtWidgets import QTabWidget
				self.view = git.run()
				self.tabs = QTabWidget()
				self.tabs.addTab(self.view, "Files")
				self.commit_view = git.make_commit()
				self.tabs.addTab(self.commit_view, "Commit")
				### test
				#self.branches_view = git.make_branches()
				#self.tabs.addTab(self.branches_view, "Branches")
				#self.status_view = git.make_status()
				#core.actionEditor().layout().addWidget(self.view)
				def p2():
					clearLayout(core.actionEditor().layout())
					core.actionEditor().layout().addWidget(self.tabs)
					#make_git_context_menu()
				#try_until_successful(p2)
				p2()
				#core.actionEditor().layout().addWidget(self.commit_view)
				#core.actionEditor().layout().addWidget(self.status_view)
				#core.actionEditor().layout().addWidget(self.asdf)
				#self.asdf.show()
				#debug_prompt(locals())
				#self.view.show()
				#self.view.hide()
				#self.view.show()
				#self.view.repaint()


			#try_until_successful(make_git)

			def diff_between_objects(a, b):
				for k in dir(a):
					if getattr(a, k) != getattr(b, k):
						print(k, getattr(a, k), getattr(b, k))	
			


			def preview():
				import os
				#print("preview")
				core = self.core
				fileName = core.formWindowManager().activeFormWindow().fileName()
				path = os.path.dirname(inspect.getfile(bsstudio))
				path_import = "import sys\nsys.path.insert(0, '"+path+"')"

				#cmd = 'ipython --profile=collection --matplotlib=qt5 -c "'+path_import+'\nimport bsstudio\nbsstudio.load(\\"'+fileName+'\\", verbose=True)"'
				#cmd = 'bsui -c "'+path_import+'\nimport bsstudio\nbsstudio.load(\\"'+fileName+'\\", False, verbose=True)"'
				cmd = 'bsui -c "import bsstudio\nbsstudio.load(\\"'+fileName+'\\", False, verbose=True)"'
				#print(core.formWindowManager().children())
				os.system(cmd + " &")
				
				#debug_prompt(locals())
				#main.menuWidget().addMenu(git.git_menu())
				#make_git()
				
				#diff_between_objects(orig, self.win)
				#core.formWindowManager().actionVerticalLayout()
				#QDesignerFormWindowInterface(None,Qt.Dialog)
				#for c in core.children():
				#	print(c.dumpObjectInfo())

				#open_template()
				#p = core.formWindowManager().findChildren(QObject)

	
				
			p = core.formWindowManager().findChild(QAction, "__qt_default_preview_action")
			p.triggered.disconnect()
			p.triggered.connect(preview)
			core.formWindowManager().formWindowAdded.connect(preview)
			#print("make git", p)
			#print(p.parent())
			#print(p.parent().children())
			#print(p.parent().parent().children())
			#print(p.parent().parent().parent())
			#print(p.parent().parent().parent().children())
			#debug_prompt(locals())
			
			
			core_initialized = True
			#make_git()

		def initialize(self, core):
			if self.initialized:
				return
			self.core = core
			self.init_core()

			self.manager = core.extensionManager()
			if self.manager:
				factory = EditCodeTaskMenuFactory(parent=self.manager)
				self.manager.registerExtensions(factory,'org.qt-project.Qt.Designer.TaskMenu')	
			self.initialized = True


	class Plugin(PyBSPlugin):
		def __init__(self):
			super(Plugin, self).__init__(cls)


	return Plugin

pCodeButton = plugin_factory(CodeButton)
#pREButton = plugin_factory(REButton)
pTextUpdate = plugin_factory(TextUpdate)
pMplWidget = plugin_factory(MplWidget)
pScan1DButton = plugin_factory(Scan1DButton)
pEmbedFrame = plugin_factory(EmbedFrame)
pLineInput = plugin_factory(LineInput)
pRECustomPlan = plugin_factory(RECustomPlan)
pCodeContainter = plugin_factory(CodeContainer, is_container=True)
pDataBrowser = plugin_factory(DataBrowser)
pOphydProperties = plugin_factory(OphydProperties)
pOpenWindowButton = plugin_factory(OpenWindowButton)
#pOpenWindowButton2 = plugin_factory(OpenWindowButton2)
pArrayImage = plugin_factory(ArrayImage)
pBooleanLED = plugin_factory(BooleanLED)
pBarUpdate = plugin_factory(BarUpdate)



import cola
import cola.app
import cola.main
from functools import partial
from cola.qtutils import add_action
from cola.i18n import N_
from cola.widgets.browse import worktree_browser
from cola.widgets.toolbar import ToolBar
from cola.widgets.commitmsg import CommitMessageEditor
from cola.widgets.status import StatusWidget
from cola.widgets.branch import BranchesWidget
from PyQt5.QtWidgets import QApplication, QMenu
from PyQt5.QtCore import QCoreApplication

import cola.models.prefs
cola.models.prefs.Defaults.editor="designer"

#cola.app.initialize()
#args = cola.main.parse_args(["cola"])
#context = cola.app.application_init(args)

actions = []

def make_menu(menu):
	for a in actions:
		action = menu.addAction(a)
		#mainThread = QCoreApplication.instance().thread()
		#action.moveToThread(mainThread)

context = None
def cmd_browse(args):
	from cola.widgets.browse import worktree_browser

	#context = cola.app.application_init(args)
	global context
	context = new_context(args)
	#view = worktree_browser(context, show=False, update=False, settings=args.settings)
	view = worktree_browser(context, show=False, update=False)
	#return cola.app.application_run(context, view)
	cola.app.initialize_view(context, view)
	cola.app.default_start(context, view)
	def open_repo(context):
		cola.guicmds.open_repo(context)
		
	view.open_repo_action = add_action(
		view, N_('Open...'), partial(cola.guicmds.open_repo, context))
	#	view, N_('Open...'), partial(open_repo, context))
	view.open_repo_action.triggered.connect(view.tree.action_refresh.trigger)
	actions.append(view.open_repo_action)
	#view.tree.update_actions()
	view.tree.action_refresh.trigger()
	#view.tree.model().refresh()
	view.refresh()
	#view.open_repo_action.trigger()
	#context.app.start()
	return view

def run():
	#app = QApplication.instance()
	args = cola.main.parse_args(["cola"])

	return cmd_browse(args)

def make_commit():
	args = cola.main.parse_args(["cola"])
	#context = new_context(args)
	global context
	#context = cola.app.application_init(args)
	commit_editor = CommitMessageEditor(context, None)
	cola.app.initialize_view(context, commit_editor)
	cola.app.default_start(context, commit_editor)
	#commit_editor.show()
	#context.app.start()
	return commit_editor

def new_context(args):
	"""Create top-level ApplicationContext objects"""
	context = cola.app.ApplicationContext(args)
	#context.settings = args.settings or cola.settings.Settings.read()
	#context.settings = cola.settings.Settings.read()
	#context.settings = Settings(verify=cola.git.is_work_tree)
	context.settings = cola.settings.Settings()
	context.settings.load()
	context.git = cola.app.git.create()
	context.cfg = cola.app.gitcfg.create(context)
	context.fsmonitor = cola.app.fsmonitor.create(context)
	context.selection = cola.app.selection.create()
	context.model = cola.app.main.create(context)
	#cola.guicmds.install()
	#cola.icons.install(args.icon_themes)
	cola.icons.install(cola.app.get_icon_themes(context))
	#toolbar = ToolBar.create(context, "asdf")
	#toolbar.show()
	#context.app = cola.app.new_application(context, args)
	#context.timer = cola.app.Timer()

	return context

def make_branches():
	args = cola.main.parse_args(["cola"])
	#context = cola.app.application_init(args)
	#cola.app.process_args(args)
	#context = cola.app.new_context(args)
	context = new_context(args)
	#context = cola.app.ApplicationContext(args)
	#context.git = cola.git.create()
	#context.model = cola.models.main.create(context)
	commit_editor = BranchesWidget(context, None)
	cola.app.initialize_view(context, commit_editor)
	#cola.app.default_start(context, commit_editor)
	#commit_editor.show()
	#context.app.start()
	return commit_editor



def make_status():
	class TB:
		def add_corner_widget(self, val):
			pass
	titlebar = TB()
	status_editor = StatusWidget(context, titlebar, None)
	status_editor.show()
	return status_editor
	
	

if __name__=="__main__":
	app1 = QApplication([])
	#commit_editor = make_commit()
	branches_editor = make_branches()
	#status_editor = make_status()
	#print(context.model)
	view = run()
	#view.show()
	app1.exec_()
	print("here")



from PyQt5 import QtCore

class WorkerSignals(QtCore.QObject):
	trigger = QtCore.pyqtSignal(object, list)


class Worker(QtCore.QRunnable):
	signals = WorkerSignals()
	def __init__(self, fn, *args, **kwargs):
		super(Worker, self).__init__()
		self.fn = fn
		self.args = args
		self.kwargs = kwargs

	@QtCore.pyqtSlot()
	def run(self):
		self.fn(*self.args, **self.kwargs)
		#self.finished.emit()





from PyQt5 import uic, QtWidgets, QtCore, QtGui
import sys
#from .widgets.REButton import Worker
from .worker import Worker
import typing
from .widgets import BaseWidget
#from .widgets import all_bss_widgets
import threading
import logging

def getMainWindow() -> typing.Union[QtWidgets.QMainWindow, None]:
	# Global function to find the (open) QMainWindow in application
	app = QtWidgets.QApplication.instance()
	for widget in app.topLevelWidgets():
		#if issubclass(widget.__class__, QtWidgets.QMainWindow):
		#if widget.__class__.__name__ == 'MainWindow':
		if isinstance(widget, MainWindow):
			return widget
	return None

def isMainWindow(w):
	return isinstance(w, MainWindow)

#def ui():
#	return getMainWindow()

def getWidgetById(id):
	for w in all_bss_widgets:
		if w.id == id:
			return w
	return None

mainWindow = None

def create_main_window(f):
	#f = "/home/bsobhani/bsw/bss_test9.ui"
	#class MainWindow(QtWidgets.QMainWindow):
	global MainWindow
	class MainWindow(*uic.loadUiType(f)):
		def __init__(self, parent=None):
			self.isLoaded = False
			super().__init__(parent)
			self.uiFilePath = f

			self.setupUi(self)
			#self.ui = uic.loadUi(f)
			#self.worker = Worker(self.ui.show)
			self.worker = Worker(self.show)

			def call_func(func, params):
				func(*params)

			self.threadpool = QtCore.QThreadPool(self)
			self.threadpool.start(self.worker)
			#self.ui.show()
			self.worker.signals.trigger.connect(call_func)
			self.isLoaded = True

		def mousePressEvent(self, event):
			focused_widget = QtWidgets.QApplication.focusWidget()
			if issubclass(focused_widget.__class__, QtWidgets.QLineEdit):
				focused_widget.clearFocus()
			QtWidgets.QWidget.mousePressEvent(self, event)	

		def closeEvent(self, evt):
			print("close event")
			#for child in self.children():
			for child in self.findChildren(QtWidgets.QWidget):
				try:
					child.close()
				except:
					None
			self.deleteLater()

	global mainWindow
	mainWindow = MainWindow()

		

def load(f, noexec=False, verbose=False):
	import os
	log_dir = os.environ.get("BSSTUDIO_LOG_FILE_NAME")
	if log_dir is None:
		log_dir = "log"
	if verbose:
		logging.basicConfig(level=logging.WARN, format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", datefmt='%Y-%m-%d %H:%M:%S')
	else:
		logging.basicConfig(filename=log_dir, filemode='a', level=logging.WARN, format="%(asctime)s:%(levelname)s:%(name)s:%(message)s", datefmt='%Y-%m-%d %H:%M:%S')
	app = QtWidgets.QApplication.instance() # checks if QApplication already exists 
	if not app: # create QApplication if it doesnt exist 
		app = QtWidgets.QApplication(sys.argv)
	#app = QtWidgets.QApplication([])


	#mainWindow = MainWindow(f)
	global mainWindow
	create_main_window(f)
	mainWindow.show()
	widgets = app.allWidgets()
	for w in widgets:
		if issubclass(w.__class__, BaseWidget):
			w.resume_widget()
	#app.exec_()
	if not noexec:
		app.exec_()
	return app



from .Base import BaseWidget
from . import CodeContainer
from .REButton import makeProperty
from PyQt5.QtCore import QDateTime
from PyQt5.Qt import Qt
from PyQt5.QtWidgets import QDateTimeEdit
from PyQt5.QtWidgets import QMenu, QAction
from PyQt5.QtWidgets import QAbstractItemView
from PyQt5.QtWidgets import QScrollArea
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtWidgets import QWidget, QDialog
from PyQt5.QtWidgets import QListWidget, QTableWidget, QTableWidgetItem, QFrame, QVBoxLayout, QLabel, QPushButton
from bsstudio.functions import widgetValue, plotHeader
from collections.abc import Iterable
import time
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class ScrollMessageBox(QDialog):
	def __init__(self, *args, **kwargs):
		QMessageBox.__init__(self, *args, **kwargs)
		#scroll = QScrollArea(self)
		#scroll.setWidgetResizable(True)
		layout = QVBoxLayout()
		self.content = QLabel(self)
		self.content.setWordWrap(True)
		layout.addWidget(self.content)
		self.setLayout(layout)
		#scroll.setWidget(self.content)
		#self.setStyleSheet("QScrollArea{min-width:300 px; min-height: 400px}")

class DataTableWidgetItem(QTableWidgetItem):
	def __lt__(self, other):
		try:
			return float(self.text()) < float(other.text())
		except:
			return self.text() < other.text() 	

class DataBrowser(CodeContainer):
	def __init__(self, parent):
		super().__init__(parent)
		self._db = ""
		#self._tableColumns = '["time"]'
		self._tableColumns = ''
		self.dbObj = None
		self._dbKwargs = "{}"
		self._plots = "[]"
		self._plotArgsList = "[[]]"
		self._plotKwargsList = "[{}]"
		layout = QVBoxLayout()
		#self.listWidget = QListWidget()
		self.listWidget = QTableWidget()
		now = QDateTime.currentDateTime()
		self.startDateTime = QDateTimeEdit(now.addMonths(-6))
		self.endDateTime = QDateTimeEdit(now)
		self.loadScansButton = QPushButton("Load Scans")
		layout.addWidget(self.startDateTime)
		layout.addWidget(self.endDateTime)
		layout.addWidget(self.loadScansButton)
		layout.addWidget(self.listWidget)
		self.setLayout(layout)

		self.loadScansButton.clicked.connect(self.__updateTable)
		self.listWidget.itemSelectionChanged.connect(self.__replot)

		self.setContextMenuPolicy(Qt.CustomContextMenu)
		self.customContextMenuRequested.connect(self.showMenu)

	def showMenu(self,event):
		menu = QMenu()
		#clear_action = menu.addAction("Clear Selection", self)
		action1 = QAction("Info", self)
		#clear_action = menu.addAction("Clear Selection", self)
		clear_action = menu.addAction(action1)
		#action = menu.exec_(self.mapToGlobal(event.pos()))
		action = menu.exec_(self.mapToGlobal(event))
		print(action.text())
		if action.text() == "Info":
			messageBox = ScrollMessageBox(self)
			messageBox.content.setText(str(self.dbObj[self.currentUid()].start))
			messageBox.show()
			#print("Information")
			#print(self.dbObj[self.currentUid()].start)

	def __updateTable(self):
		self.runCode()
		self._updateTable()
		#self.updateTable(self.dbObj, {})


	def __replot(self):
		self.runCode()
		self._replot()

	def updateTable(self, db, dbKwargs):
		#self.listWidget.clear()
		since = self.startDateTime.dateTime().toString("yyyy-MM-dd HH:mm:ss")
		until = self.endDateTime.dateTime().toString("yyyy-MM-dd HH:mm:ss")
		results = list(db(since=since, until=until, **dbKwargs))
		self.listWidget.setRowCount(len(results))
		self.listWidget.setSortingEnabled(True)
		self.listWidget.setSelectionBehavior(QAbstractItemView.SelectRows)
		col_set=set([])
		for r in results:
			cols = [k for k,v in r.start.items() if not isinstance(v,Iterable) or type(v)==str]
			col_set.update(cols)	
		cols = list(col_set)

 

		self.listWidget.setColumnCount(len(cols))
		self.listWidget.setHorizontalHeaderLabels(cols)
		for i in range(len(results)):
			r = results[i]
			logger.info("len results: "+str(len(results)) +", i:"+str(i))
			#self.listWidget.addItem(r.start['uid'])
			for j in range(len(cols)):
				#item = QTableWidgetItem(r.start['uid'])
				logger.info(str(i)+" "+ str(j) + " "+str(cols[j]))
				try:
					field = r.start[cols[j]]
				except KeyError:
					field = None
				#print(str(i)+" "+ str(j) + " "+str(cols[j]))
				#print(r.start[cols[j]])
				#item = QTableWidgetItem(str(r.start[cols[j]]))
				item = DataTableWidgetItem(str(field))
				self.listWidget.setItem(i,j,item)

		if self.tableColumns!="":
			#for colName in eval(self.tableColumns):
			#	self.listWidget.hideColumn(self.findHorizontalHeaderIndex(colName))
			tableColumns = eval(self.tableColumns)
			for i in range(self.listWidget.columnCount()):
				if self.listWidget.horizontalHeaderItem(i).text() not in tableColumns:
					logger.info("hiding column "+self.listWidget.horizontalHeaderItem(i).text())
					self.listWidget.hideColumn(i)
				else:
					self.listWidget.showColumn(i)
					

	def findHorizontalHeaderIndex(self, key):
		logger.info("horizonal header count: "+str(self.listWidget.columnCount()))
		for i in range(self.listWidget.columnCount()):
			if self.listWidget.horizontalHeaderItem(i).text()==key:
				logger.info(key + " found")
				return i
		return None


	def currentUid(self):
		uids = self.currentUids()
		if len(uids)==0:
			return None
		return uids[0]

	def currentUids(self):
		#rows = self.listWidget.selectedItems()
		uid_col = self.findHorizontalHeaderIndex("uid")
		#uids = [item.text() for item in rows if item.column()==uid_col]
		rows = [item.row() for item in self.listWidget.selectedItems()]
		rows = list(set(rows))
		uids = [self.listWidget.item(row, uid_col).text() for row in rows]
		return uids
		

	def startData(self,key):
		if self.currentUid() is None:
			return []
		return self.dbObj[self.currentUid()].start[key]
		

	def replotUid(self, plots, db, uid):
		logger.info("replot uid: "+uid)
		if uid is None:
			return
		if plots is None:
			return
		for p in plots:
			if not hasattr(p, "ax"):
				p._LivePlot__setup()
		
		for p in plots:
			plotHeader(p, db[uid])
			p.ax.figure.tight_layout()
		

	def replot(self, plots, db):
		for uid in self.currentUids():
			self.replotUid(plots, db, uid)

	def default_code(self):
		return """
			if 1==1:
				ui = self.ui
				from functools import partial
				from bsstudio.functions import widgetValue
				from bsstudio.functions import makeLivePlots 
				db = widgetValue(eval(self.db))
				self.dbObj = db
				#self.startDateTime.dateTimeChanged.connect(partial(self.updateTable, db))
				#self.endDateTime.dateTimeChanged.connect(partial(self.updateTable, db))
				plots = eval(self.plots)
				for plot in plots:
					plot.canvas.ax.clear()
				try:
					plotArgsList = widgetValue(eval(self.plotArgsList))
				except TypeError:
					print(self.plotArgsList)
					print("databrowser plotargslist exception")
					plotArgsList = None
				plotKwargsList = eval(self.plotKwargsList)
				dbKwargs = widgetValue(eval(self.dbKwargs))
				livePlots = makeLivePlots(plots, plotArgsList, plotKwargsList)
				#self.replot(plots, db)
				#self.updateTable(db, dbKwargs)
				#self.listWidget.currentTextChanged.connect(partial(self.replot, plots, db))
				self._replot = partial(self.replot, livePlots, db)
				self._updateTable = partial(self.updateTable, db, dbKwargs)
				#for plot in plots:
				#	plot.canvas.update()
				#	plot.update()
				#	plot.canvas.draw()
				
				
			"""[1:]

	db = makeProperty("db")
	dbKwargs = makeProperty("dbKwargs")
	plots = makeProperty("plots")
	tableColumns = makeProperty("tableColumns")
	plotArgsList = makeProperty("plotArgsList")
	plotKwargsList = makeProperty("plotKwargsList")



from PyQt5 import QtDesigner, QtGui, QtWidgets, QtCore
#from qtpy.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton, QPlainTextEdit
#from qtpy.QtDesigner import QExtensionFactory
from PyQt5.QtDesigner import QExtensionFactory
from PyQt5.QtCore import pyqtProperty as Property
from ophyd.sim import det
import inspect
from itertools import dropwhile
import textwrap
from .CodeObject import CodeObject
from functools import partial

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

class CodeButton(QPushButton, CodeObject):
	def __init__(self, parent=None):
		self.parent = parent
		QPushButton.__init__(self, parent)
		CodeObject.__init__(self, parent)
		self.clicked.connect(self.runCode)
		#self._useThreading = True
		self.threadType = "qthread"

	def default_code(self):

		code_string = """
		ui = self.ui
		print(self.objectName()+" pressed...")
		"""[1:]
		return code_string

	def runCode(self):
		if self.threadpool.waitForDone(0):
			CodeObject.runCode(self)
		else:
			logger.info("Thread still running for CodeButton")
	



from PyQt5 import QtDesigner, QtGui, QtWidgets, QtCore
#from qtpy.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton, QPlainTextEdit, QFrame
#from qtpy.QtDesigner import QExtensionFactory
from PyQt5.QtDesigner import QExtensionFactory
from PyQt5.QtCore import pyqtProperty as Property
from PyQt5.QtCore import QThread
from PyQt5.QtCore import QReadWriteLock
from PyQt5.Qt import Qt
from ophyd.sim import det
import inspect
from itertools import dropwhile
import textwrap
from .CodeObject import CodeObject
from ophyd.ophydobj import OphydObject
import logging
from ..worker import Worker
import time
import threading
import sip
from .REButton import makeProperty

logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.WARN)

main_thread = threading.current_thread()

def isOphyd(obj):
	return issubclass(obj, OphydObject)

class WorkerThread(QThread):
	lock = QReadWriteLock()
	cancelled = False
	def cancel(self):
		WorkerThread.cancelled = True
	def resume(self):
		WorkerThread.cancelled = False
	def setFunc(self, func):
		self.func = func
	def run(self):
		self.func()	

class TextUpdateBase(CodeObject):
	def __init__(self, parent=None,*,sig=""):
		self.parent = parent
		#super().__init__(parent)
		#QLabel.__init__(self, parent)
		CodeObject.__init__(self, parent)
		self.updatePeriod_ = 1500
		self._updatePeriod = "1500"
		self._source = ""
		self.source = sig
		self._useThreading = False
		self.threadpool.setMaxThreadCount(1)
		#self.worker = Worker(self.start_thread)
		t0 = time.time()
		self.worker = WorkerThread(self)
		self.worker.setFunc(self.start_thread)
		self.timer = QtCore.QTimer(self)
		self.timer.setInterval(self.updatePeriod_)
		self.timer.timeout.connect(self.timeout)
		logger.info(str(time.time()-t0) +" seconds to create thread")
		self.threadMode = "qtimer"
		self.start_time = time.time()

	def timeout(self):
		t0 = time.time()
		self.runCode()
		logger.info("Timeout function duration: "+str(time.time()-t0))

	def resumeWorkerThread(self):
		WorkerThread.cancelled = False

	def start_thread(self):
		self.update()
		self.repaint()
		time.sleep(10)
		logger.info("Starting thread")
		t0 = time.time()
		WorkerThread.cancelled = False
		while not WorkerThread.cancelled and not sip.isdeleted(self):
			#if time.time()-t0>self.updatePeriod_/1000:
			if time.time()-t0>self.updatePeriod_/1000 and main_thread.isAlive():
				self.timeout()
				logger.info(self.objectName()+":runCode duration: "+str(time.time()-t0))
				logger.info("update period: "+str(self.updatePeriod_))
				time.sleep(self.updatePeriod_/10000)
				#self.worker.wait(10)
				t0 = time.time()
			if not main_thread.isAlive():
				break
		logger.info("end of thread")

	def updateText(self, val):
		if val == None:
			self.setText("unknown")
			logger.info("setting text to unknown")
		else:
			self.setText(val)
	

	def default_code(self):
		return """
			from bsstudio.functions import widgetValue, fieldValue
			ui = self.ui
			try:
				v = widgetValue(fieldValue(self, "source"))
			except:
				v=None
			if v is not None:
				v = str(v)
			self.updateText(v)
			"""[1:]

	@Property(str, designable=True)
	def source(self):
		return self._source

	@source.setter
	def source(self, val):
		self._source = val

	def pause_widget(self):
		self._paused = True

	def closeEvent(self, evt):
		logger.info("close Event")
		self.worker.cancel()
		while not self.worker.isFinished() and self.worker.isRunning():
			logger.info("worker not finished")
			time.sleep(2)
		super().closeEvent(evt)

	def resume_widget(self):
		CodeObject.resume_widget(self)
		#self.threadpool.clear()
		#self.threadpool.start(self.worker)
		t0 = time.time()
		self.updatePeriod_ = eval(self.updatePeriod)
		if self.threadMode == "qtimer":
			self.timer.setInterval(self.updatePeriod_)
			self.timer.start()
		elif self.threadMode == "qthread":
			self.worker.start()
		logger.info(str(time.time()-t0) +" seconds to start thread")

	updatePeriod = makeProperty("updatePeriod")


class TextUpdate(QLabel, TextUpdateBase):
	def __init__(self, parent=None,*,sig=""):
		self.parent = parent
		#super().__init__(parent)
		QLabel.__init__(self, parent)
		TextUpdateBase.__init__(self, parent, sig=sig)
		#self.setStyleSheet("border: 1px solid black;")
		#self.setLineWidth(1)
		#self.setAlignment(Qt.AlignHCenter)
		self.setAlignment(Qt.AlignCenter)
		self.setFrameShape(QFrame.Box)
		self.setFrameShadow(QFrame.Sunken)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import check_call
import os


class PostInstallCommand(install):
    def run(self):
        os.system("mkdir -p $CONDA_PREFIX/etc/conda/activate.d")
        os.system("mkdir -p $CONDA_PREFIX/etc/bsstudio")
        os.system('echo "from bsstudio.qtplugins import *"> $CONDA_PREFIX/etc/bsstudio/load_plugin.py')
        os.system('echo "export PYQTDESIGNERPATH=$CONDA_PREFIX/etc/bsstudio"> $CONDA_PREFIX/etc/conda/activate.d/env_vars.sh')

        #os.system("conda install --yes --file requirements.txt")
        #os.system("conda install --yes -c conda-forge/label/cf201901 pyqt")
        #os.system("conda install --yes bluesky -c lightsource2-tag")
        install.run(self)
        
setuptools.setup(
    name="bsstudio",
    version="0.0.1",
    author="Bayan Alexander Sobhani",
    author_email="bsobhani@bnl.gov",
    description="Bluesky widgets for PyQt",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    cmdclass={
        'install': PostInstallCommand,
    },
)



from PyQt5 import QtDesigner, QtGui, QtWidgets, QtCore
#from qtpy.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton, QPlainTextEdit
#from qtpy.QtDesigner import QExtensionFactory
from PyQt5.QtDesigner import QExtensionFactory
from PyQt5.QtCore import pyqtProperty as Property
from ophyd.sim import det
import inspect
from itertools import dropwhile
import textwrap
from .CodeButton import CodeButton

class REAbortButton(CodeButton):
	def default_code(self):

		return """
		ui = self.ui
		RE.abort()
		"""[1:]
	



from PyQt5 import QtDesigner, QtGui, QtWidgets, QtCore
#from qtpy.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel, QApplication, QDoubleSpinBox, QWidget, QPushButton, QPlainTextEdit, QComboBox
#from qtpy.QtDesigner import QExtensionFactory
from PyQt5.QtDesigner import QExtensionFactory
from PyQt5.QtCore import pyqtProperty as Property
from ophyd.sim import det
import inspect
from itertools import dropwhile
import textwrap
from .CodeButton import CodeButton
from .CodeObject import CodeObject
from ..worker import Worker, WorkerSignals
from IPython import get_ipython


def parseField(field):
	ip = get_ipython()
	#obj = eval(field, ip.user_ns)
	#obj = eval(field)
	obj = field
	if isinstance(obj, QComboBox):
		obj = eval(obj.currentText(), ip.user_ns)
		#obj = eval(obj.currentText())
	return obj



def makeProperty(name, propertyType=str, **kwargs):
	storageVarName = "_"+name
	def g(self):
		return eval("self."+storageVarName)

	def s(self, val):
		exec("self."+storageVarName+"=val")	
		if "notify" in kwargs.keys():
			kwargs["notify"].emit()
	
	return Property(propertyType, g, s)


	
		


class REButton(CodeButton):
	def __init__(self, parent):
		super().__init__(parent)
		self._plots = "[]"
		self._plotFields = "[[]]"
		self._plotKwargsList = "[{}]"
		self._RunEngine = "RE"
	plots = makeProperty("plots")
	plotFields = makeProperty("plotFields")
	plotKwargsList = makeProperty("plotKwargsList")
	RunEngine = makeProperty("RunEngine")


class RECustomPlan(REButton):
	def __init__(self, parent):
		super().__init__(parent)
		self._arguments = []
		self._plan = ""

	arguments = makeProperty("arguments", "QStringList")
	plan = makeProperty("plan")
	def default_code(self):
		return """
			from bsstudio.functions import widgetValue, makeLivePlots
			from bluesky.callbacks import LiveGrid
			from bluesky.plans import grid_scan
			ui = self.ui
			args = []
			kwargs = {}
			for arg in self.arguments:
				sp = arg.split("=")
				left = sp[0]
				right = widgetValue(eval(sp[-1]))
				if len(sp)==1:
					args.append(right)
				else:
					kwargs[left] = right
			RE = widgetValue(eval(self.RunEngine))
			plots = widgetValue(eval(self.plots)[:])
			plotFields = widgetValue(eval(self.plotFields)[:])
			plotKwargsList = eval(self.plotKwargsList)[:]
			livePlots = makeLivePlots(plots, plotFields, plotKwargsList)
			if livePlots is None:
				livePlots = []
			ts = [RE.subscribe(lp) for lp in livePlots]
	
			plan = widgetValue(eval(self.plan))
			#RE(plan(*args,**kwargs),LiveGrid((6, 10), 'det4'))
			RE(plan(*args,**kwargs))
			for p in plots:
				try:
					x = p[0]
				except:
					x = p
				x.canvas.draw()
			for t in ts:
				RE.unsubscribe(t)
			"""[1:]
			



class Scan1DButton(REButton):
	def __init__(self, parent):
		super().__init__(parent)
		self._motor = ""
		self._detector_list = "[]"
		self._startPosition = "-1"
		self._endPosition = "1"
		self._numSteps = "10"

	def default_code(self):

		return """
		from bsstudio.functions import isWidget, widgetValue, makeLivePlots
		ui = self.ui

		detector_list = eval(self.detectorList)[:]
		motor = eval(self.motor)
		endPosition = eval(self.endPosition)
		startPosition = eval(self.startPosition)
		numSteps = eval(self.numSteps)
		RE = eval(self.RunEngine)

		ophyd_detector_list = [widgetValue(w) for w in detector_list]

		motor = widgetValue(motor)
		endPosition = widgetValue(endPosition)
		startPosition = widgetValue(startPosition)
		numSteps = widgetValue(numSteps)
		RE = widgetValue(RE)

		plots = eval(self.plots)[:]
		plotFields = eval(self.plotFields)[:]
		plotKwargsList = eval(self.plotKwargsList)[:]
		livePlots = makeLivePlots(plots, plotFields, plotKwargsList)
		if livePlots is None:
			livePlots = []
		ts = [RE.subscribe(lp) for lp in livePlots]
	
		plan = scan(ophyd_detector_list, motor, startPosition, endPosition, numSteps)
		#Worker.signals.trigger.emit(RE, [plan])
		RE(plan)
		for t in ts:
			RE.unsubscribe(t)
		"""[1:]

	startPosition = makeProperty("startPosition")
	endPosition = makeProperty("endPosition")
	numSteps = makeProperty("numSteps")
	motor = makeProperty("motor")

	@Property(str)
	def detectorList(self):
		return self._detector_list

	@detectorList.setter
	def detectorList(self, val):
		self._detector_list = val[:]



from .CodeObject import CodeObject
from PyQt5.QtWidgets import QFrame, QWidget, QLabel, QMessageBox
from PyQt5.QtCore import QFile, QFileSelector, QUrl, QVariant, pyqtSignal, QDir
from PyQt5.QtCore import pyqtProperty as Property
from .REButton import makeProperty
from .CodeButton import CodeButton
from .Base import BaseWidget
from bsstudio.functions import openFileAsString, getTopObject
import os

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def relPath(selfPath, filePath):
	#cp = os.path.commonpath([selfPath, filePath])
	#print("rel ins:",selfPath, filePath)
	selfDir = os.path.dirname(selfPath)
	try:
		path = os.path.relpath(filePath, selfDir)
	except:
		path = ""
	#print("rel out:",path)
	return path

def absPath(selfPath, relFilePath):
	#print("selfpath", selfPath)
	#print("relfilepath",relFilePath)
	prefix = os.path.dirname(selfPath)
	ap = os.path.join(prefix, relFilePath)
	#print("abs path", ap)
	return ap

def convertPath(w, fileUrl,*,toRelative):
	val = fileUrl
	valPath = val.toLocalFile()
	self = w
	if self.windowFileName()=="":
		alert = QMessageBox(self)
		alert.setText("Current file has no name. Please save the current file first and try again.")
		alert.show()
		#self._fileName=val
		return None
	if valPath=="":
		alert = QMessageBox(self)
		alert.setText("Empty filename")
		alert.show()
		#self._fileName=val
		return None
	if self.windowFileName() is None or toRelative!=os.path.isabs(valPath):
		#self._fileName=val
		#return None
		logger.info("self.windowFileName is None")
		#return None
		return val
	if toRelative:
		rp = relPath(self.windowFileName(), valPath)
		#print("w rp", rp)
		return QUrl("file:"+rp)
	else:
		ap = absPath(self.windowFileName(), valPath)
		#print("w ap", ap)
		return QUrl("file:"+ap)



class CodeContainer(QFrame, CodeObject):
	def __init__(self, parent=None):
		super().__init__(parent)
		self.pause_widget()
	

	def default_code(self):
		return """
			ui = self.ui
			"""[1:]

	def resume_widget(self):
		#self._paused = False
		CodeObject.resume_widget(self)
		self.runCode()
		

def deleteWidgetAndChildren(w):
	w.setParent(None)
	for c in w.children():
		c.deleteLater()
	w.deleteLater()


class EmbedFrame(QFrame, CodeObject):

	fileChanged = pyqtSignal()

	def deleteSubWindow(self):
		if hasattr(self,"subWindow"):
			#self.subWindow.setParent(None)
			#for c in self.subWindow.children():
			#	c.deleteLater()
			#self.subWindow.deleteLater()
			deleteWidgetAndChildren(self.subWindow)
		else:
			logger.info("no existing subwindow")



	def __init__(self, parent=None):
		#super().__init__(parent)
		QFrame.__init__(self,parent)
		CodeObject.__init__(self,parent)
		self._fileName = QUrl()
		self._macros = []
		self._useRelativePath = True
		self.fileChanged.connect(self.updateUi)
		original = self.resizeEvent
		def resizeEvent(event):
			original(event)
			#self.updateUi()
			self.runCode()
		self.resizeEvent = resizeEvent

	def runPaused(self):
		self.updateUi()

	def getAbsPath(self):
		filename = self.fileName.toLocalFile()
		if not QDir.isAbsolutePath(filename):
			filename = absPath(self.windowFileName(), filename)
		return filename
	
	def updateUi(self):
		from PyQt5.QtWidgets import QWidget
		from PyQt5.QtWidgets import QVBoxLayout
		ui = self.ui

		from PyQt5 import uic
		import io
		#if hasattr(self,"subWindow"):
		#	self.subWindow.setParent(None)
		#	for c in self.subWindow.children():
		#		c.deleteLater()
		#	self.subWindow.deleteLater()
		self.deleteSubWindow()

		self.subWindow = QWidget(self)
		
		self.subWindow.isTopLevel = True
		filename = self.fileName.toLocalFile()
		#if filename=="" or filename==None:
		#	return
		if self.windowFileName() is None:
			return
		if not QDir.isAbsolutePath(filename):
			filename = absPath(self.windowFileName(), filename)
		self.subWindow.uiFilePath = filename
		#if filename=="":
		#	return
		fileContents = openFileAsString(filename, self.macros)
		fileObject = io.StringIO(fileContents)
		try:
			uic.loadUi(fileObject, self.subWindow)
		except:
			logger.info("Error opening file "+filename)
			return
		self.subWindow.resize(self.size())
		#print("subwindow show")
		self.subWindow.show()
		if not self._paused:
			self.resume_children()

		

	

	def default_code(self):
		return """
			from PyQt5.QtWidgets import QWidget
			from PyQt5 import uic
			import io
			ui = self.ui
			self.updateUi()
			"""[1:]

	def resume_widget(self):
		#self._paused = False
		CodeObject.resume_widget(self)
		self.runCode()
		

	#fileName = makeProperty("fileName", QUrl, notify=fileChanged)
	macros = makeProperty("macros", "QStringList")
	@Property("QUrl")
	def fileName(self):
		return self._fileName

	@fileName.setter
	def fileName(self, val):
		"""
		valPath = val.toLocalFile()
		if self.windowFileName() is None:
			self._fileName=val
			return
		rp = relPath(self.windowFileName(), valPath)
		ap = absPath(self.windowFileName(), rp)
		if self._useRelativePath:
			self._fileName=QUrl("file:"+rp)
		else:
			self._fileName=QUrl("file:"+ap)
		self.fileChanged.emit()
		"""
		
			
		path = convertPath(self, val, toRelative=self._useRelativePath)
		if path is not None:
			#print("path",path)
			self._fileName=path
			self.fileChanged.emit()
		return
			
			

	@Property(bool)
	def useRelativePath(self):
		return self._useRelativePath

	@useRelativePath.setter
	def useRelativePath(self, val):
		self._useRelativePath = val
		self.fileName = self._fileName


	#useRelativePath = makeProperty("useRelativePath", bool)


class OpenWindowButtonOld(CodeButton):

	def __init__(self, parent=None):
		super().__init__(parent)
		self._fileName = QUrl()
		self._macros = []
		self._useRelativePath = True
		self._useThreading = False
	
	def default_code(self):
		return """
			from PyQt5.QtWidgets import QDialog
			from PyQt5 import uic, QtCore
			from PyQt5.QtCore import QDir
			from bsstudio.functions import openFileAsString
			from bsstudio.widgets.embedframe import absPath
			import io
			ui = self.ui
			self.subWindow = QDialog(self)
			self.subWindow.setAttribute(QtCore.Qt.WA_DeleteOnClose)

			self.subWindow.isTopLevel = True
			filename = self.fileName.toLocalFile()
			if not QDir.isAbsolutePath(filename):
				filename = absPath(self.windowFileName(), filename)
			fileContents = openFileAsString(filename, self.macros)
			fileObject = io.StringIO(fileContents)
			uic.loadUi(fileObject, self.subWindow)
			#self.resize(self.subWindow.size())
			self.subWindow.show()
			
			self.resume_children()
			"""[1:]

	def resume_widget(self):
		self._paused = False
	

	@Property("QUrl")
	def fileName(self):
		return self._fileName

	@fileName.setter
	def fileName(self, val):
		path = convertPath(self, val, toRelative=self._useRelativePath)
		if path is not None:
			self._fileName=path

	@Property(bool)
	def useRelativePath(self):
		return self._useRelativePath

	@useRelativePath.setter
	def useRelativePath(self, val):
		self._useRelativePath = val
		self.fileName = self._fileName

		

	#fileName = makeProperty("fileName", QUrl)
	macros = makeProperty("macros", "QStringList")

class OpenWindowButton(CodeButton):


	def deleteSubWindow(self):
		if hasattr(self,"subWindow"):
			deleteWidgetAndChildren(self.subWindow)
		else:
			logger.info("no existing subwindow")




	def __init__(self, parent=None):
		super().__init__(parent)
		self._fileName = QUrl()
		self._macros = []
		self._useRelativePath = True
		self._useThreading = False
	
	def default_code(self):
		return """
			from PyQt5.QtWidgets import QDialog, QFrame, QWidget, QApplication
			from PyQt5 import uic, QtCore
			from PyQt5.QtCore import QDir
			from PyQt5.Qt import Qt
			from bsstudio.functions import openFileAsString
			from bsstudio.widgets.embedframe import absPath, EmbedFrame
			import time
			import io
			ui = self.ui
			#if hasattr(self, "subWindow"):
			#	print("already has subwindow")
			#	self.subWindow.close()
			#self.deleteSubWindow()
			#self.subWindow = QDialog(self)
			self.subWindow = EmbedFrame(self)
			self.subWindow.setWindowFlags(Qt.Window)
			#self.subWindow = QDialog(self)
			self.subWindow.setAttribute(QtCore.Qt.WA_DeleteOnClose)

			self.subWindow.isTopLevel = True
			filename = self.fileName.toLocalFile()
			if not QDir.isAbsolutePath(filename):
				filename = absPath(self.windowFileName(), filename)
			self.subWindow.uiFilePath = filename
			fileContents = openFileAsString(filename, self.macros)
			fileObject = io.StringIO(fileContents)
			uic.loadUi(fileObject, self.subWindow)
			#self.resize(self.subWindow.size())
			self.subWindow.show()
			self.subWindow.update()
			self.subWindow.repaint()
			#QApplication.instance().processEvents()
			self.resume_children()
			"""[1:]

	def resume_widget(self):
		self._paused = False
	

	@Property("QUrl")
	def fileName(self):
		return self._fileName

	@fileName.setter
	def fileName(self, val):
		path = convertPath(self, val, toRelative=self._useRelativePath)
		if path is not None:
			self._fileName=path

	@Property(bool)
	def useRelativePath(self):
		return self._useRelativePath

	@useRelativePath.setter
	def useRelativePath(self, val):
		self._useRelativePath = val
		self.fileName = self._fileName

		

	#fileName = makeProperty("fileName", QUrl)
	macros = makeProperty("macros", "QStringList")




from PyQt5.QtWidgets import QWidget, QComboBox, QSpinBox, QDoubleSpinBox, QCheckBox, QLineEdit
from bluesky.callbacks import LivePlot, LiveGrid
import matplotlib.pyplot as plt
from collections import Iterable
import time
import re
from IPython import get_ipython
from .lib.pydollarmacro import pydollarmacro

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)


class dotdict(dict):
	__getattr__ = dict.get
	__setattr__ = dict.__setitem__
	__delattr__ = dict.__delitem__

def getTopObject(w):
	from .window import isMainWindow
	obj = w.parentWidget()
	while True:
		if hasattr(obj, "isTopLevel"):
			if obj.isTopLevel==True:
				#return obj
				break
		if isMainWindow(obj):
			while not obj.isLoaded:
				time.sleep(1)
				logger.info("sleeping for 1 second")
			#print(obj.findChildren(QWidget))
			#return obj
			break
		obj = obj.parentWidget()
	return obj


def makeUiFunction(self):
	def ui():
		obj = getTopObject(self)
		children = obj.findChildren(QWidget)
		d = {c.objectName(): c for c in children}
		d["parent"] = obj.parent
		def parentUi():
			return obj.parent.ui()
		d["parentUi"] = parentUi
		

		d = dotdict(d)
		return d
	return ui

def defaultValueField(w):
	if isinstance(w, QComboBox):
		return "currentText"
	if isinstance(w, QCheckBox):
		return "isChecked"
	if isinstance(w, QSpinBox):
		return "value"
	if isinstance(w, QDoubleSpinBox):
		return "value"
	if issubclass(w.__class__, QLineEdit):
		return "text"

	return None

def fieldValueAsString_(w, field):
	prop = w.property(field)
	if prop == None:
		return str(getattr(w, field)())
	return str(w.property(field))

def fieldValueAsString(w, field):
	try:
		return fieldValueAsString_(w,field)
	except:
		logger.warn("unable to get value for field "+field+" as string")
		return None

def evalInNs(w, cmd):
	try:
		ip = get_ipython()
		ns = ip.user_ns.copy()
		ns['self'] = w
		ui = makeUiFunction(w)
		ns["ui"] = ui
		logger.info("evaluating command in namespace: "+cmd)
		return eval(cmd, ns)

	except:
		logger.warn("unable to evaluate command "+str(cmd)+" in namespace "+str(ns.keys()))
		return None



def fieldValue(w, field):
	return evalInNs(w, fieldValueAsString(w, field))

def comboBoxValue(w):
	key = w.currentText()
	prop = w.property(key)
	if prop == None:
		key = "currentText"
	return fieldValue(w, key)
	
	
def widgetValueString(self, w_string, continuous=True):
	#ui = makeUiFunction(w)
	w = evalInNs(self, w_string)
	logger.debug("w in widgetValueString", w)
	if type(w) is list:
		return [widgetValueString(x, continuous) for x in w]
	if not isWidget(w):
		return w_string
	return widgetValue(w, continuous, asString=True)


def widgetValue(w, continuous=True,*,asString=False):
	if type(w) is list:
		return [widgetValue(x, continuous) for x in w]
	if type(w) is dict:
		return {k : widgetValue(w[k], continuous) for k in w.keys()}
	if not isWidget(w):
		return w
	prop = w.property("valueField")
	if prop == None:
		prop = defaultValueField(w)
	if isinstance(w, QComboBox):
		wv = comboBoxValue(w)
	else:
		wv = fieldValue(w, prop)
	wv_string = fieldValueAsString(w, prop)
	if not isWidget(wv) and asString:
		return wv_string
	if continuous:
		return widgetValue(wv, True, asString=asString)
	return wv



def isWidget(obj):
	return issubclass(obj.__class__, QWidget)
	
def makeLivePlots(plots, plotFields, plotKwargsList):
	logger.info("plotFields"+str(plotFields))
	if plotFields is None or plotFields == [[]]:
		return None
	plotKwargsList = plotKwargsList + [{}]*(len(plots)-len(plotKwargsList))
	livePlots = []
	for i in range(len(plots)):
		plot_tuple = plots[i]
		if not isinstance(plot_tuple, Iterable):
			plot_tuple = (plot_tuple, "LivePlot")
		plotWidget,cls = plot_tuple
		if cls == "LiveGrid":
			plotWidget.canvas.wipe()
		p = plotFields[i]
		plotKwargs = plotKwargsList[i]
		plotKwargs["ax"] = plotWidget.canvas.ax
		lp = eval(cls)(*p, **plotKwargs)
		livePlots.append(lp)
	return livePlots

def plotHeader(livePlot, header):
	livePlot.start(header.start)
	for e in header.events():
		livePlot.event(e)

def openFileAsString(filename, macros=[]):
	try:
		fileContents = open(filename).read()
	except:
		logger.error("Read error")
		return

	macro_dict = {}
	for m in macros:
		left, right = m.split(":")
		macro_dict[left] = right
		#fileContents = fileContents.replace("$("+left+")", right)
		#fileContents = fileContents.replace("$("+left+")", right)
	fileContents = pydollarmacro.subst_str_all(fileContents, macro_dict)
	return fileContents





from .REButton import REButton, Scan1DButton, parseField, RECustomPlan
from .CodeButton import CodeButton
from .TextUpdate import TextUpdate
from .mplwidget import MplWidget
from .Base import all_bss_widgets, BaseWidget
from .embedframe import EmbedFrame, CodeContainer, OpenWindowButton
from .lineinput import LineInput
from .databrowser import DataBrowser
from .arrayimage import ArrayImage
from .ophydproperties import OphydProperties
from .booleanled import BooleanLED
from .barupdate import BarUpdate
